"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.serializeUint = exports.serializeLayout = void 0;
const layout_1 = require("./layout");
const utils_1 = require("./utils");
function serializeLayout(layout, data, encoded, offset = 0) {
    let ret = encoded ?? new Uint8Array(calcLayoutSize(layout, data));
    for (let i = 0; i < layout.length; ++i)
        offset = serializeLayoutItem(layout[i], data[layout[i].name], ret, offset);
    return encoded === undefined ? ret : offset;
}
exports.serializeLayout = serializeLayout;
const calcLayoutSize = (layout, data) => layout.reduce((acc, item) => {
    switch (item.binary) {
        case "object": {
            return acc + calcLayoutSize(item.layout, data[item.name]);
        }
        case "array": {
            if (item.lengthSize !== undefined)
                acc += item.lengthSize;
            const narrowedData = data[item.name];
            for (let i = 0; i < narrowedData.length; ++i)
                acc += calcLayoutSize(item.layout, narrowedData[i]);
            return acc;
        }
        case "bytes": {
            if (item.custom !== undefined) {
                if (item.custom instanceof Uint8Array)
                    return acc + item.custom.length;
                else if (item.custom.from instanceof Uint8Array)
                    return acc + item.custom.from.length;
            }
            item = item;
            if ("size" in item && item.size !== undefined)
                return acc + item.size;
            if (item.lengthSize !== undefined)
                acc += item.lengthSize;
            return acc + ((item.custom !== undefined)
                ? item.custom.from(data[item.name])
                : data[item.name]).length;
        }
        case "uint": {
            return acc + item.size;
        }
    }
}, 0);
//Wormhole uses big endian by default for all uints
//endianess can be easily added to UintLayout items if necessary
function serializeUint(encoded, offset, val, bytes) {
    if (val < 0 || (typeof val === "number" && !Number.isInteger(val)))
        throw new Error(`Value ${val} is not an unsigned integer`);
    if (bytes > layout_1.numberMaxSize && typeof val === "number" && val >= 2 ** (layout_1.numberMaxSize * 8))
        throw new Error(`Value ${val} is too large to be safely converted into an integer`);
    if (val >= 2n ** BigInt(bytes * 8))
        throw new Error(`Value ${val} is too large for ${bytes} bytes`);
    //big endian byte order
    for (let i = 0; i < bytes; ++i)
        encoded[offset + i] = Number((BigInt(val) >> BigInt(8 * (bytes - i - 1)) & 0xffn));
    return offset + bytes;
}
exports.serializeUint = serializeUint;
function serializeLayoutItem(item, data, encoded, offset) {
    try {
        switch (item.binary) {
            case "object": {
                for (const i of item.layout)
                    offset = serializeLayoutItem(i, data[i.name], encoded, offset);
                break;
            }
            case "array": {
                if (item.lengthSize !== undefined)
                    offset = serializeUint(encoded, offset, data.length, item.lengthSize);
                for (let i = 0; i < data.length; ++i)
                    offset = serializeLayout(item.layout, data[i], encoded, offset);
                break;
            }
            case "bytes": {
                const value = (() => {
                    if (item.custom !== undefined) {
                        if (item.custom instanceof Uint8Array) {
                            if (!item?.omit)
                                (0, utils_1.checkUint8ArrayDeeplyEqual)(item.custom, data);
                            return item.custom;
                        }
                        if (item.custom.from instanceof Uint8Array)
                            //no proper way to deeply check equality of item.custom.to and data in JS
                            return item.custom.from;
                    }
                    item = item;
                    const ret = item.custom !== undefined ? item.custom.from(data) : data;
                    if ("size" in item && item.size !== undefined)
                        (0, utils_1.checkUint8ArraySize)(ret, item.size);
                    else if (item.lengthSize !== undefined)
                        offset = serializeUint(encoded, offset, ret.length, item.lengthSize);
                    return ret;
                })();
                encoded.set(value, offset);
                offset += value.length;
                break;
            }
            case "uint": {
                const value = (() => {
                    if (item.custom !== undefined) {
                        if (typeof item.custom == "number" || typeof item.custom === "bigint") {
                            if (!item?.omit)
                                (0, utils_1.checkUintEquals)(item.custom, data);
                            return item.custom;
                        }
                        if (typeof item.custom.from == "number" || typeof item.custom.from === "bigint")
                            //no proper way to deeply check equality of item.custom.to and data in JS
                            return item.custom.from;
                    }
                    return item.custom !== undefined ? item.custom.from(data) : data;
                })();
                offset = serializeUint(encoded, offset, value, item.size);
                break;
            }
        }
    }
    catch (e) {
        e.message = `when serializing item '${item.name}': ${e.message}`;
        throw e;
    }
    return offset;
}
;
//# sourceMappingURL=serialize.js.map