import { numberMaxSize, } from "./layout";
import { checkUint8ArrayDeeplyEqual, checkUintEquals } from "./utils";
export function deserializeLayout(layout, encoded, offset = 0, consumeAll = true) {
    const [decoded, finalOffset] = internalDeserializeLayout(layout, encoded, offset);
    if (consumeAll && finalOffset !== encoded.length)
        throw new Error(`encoded data is longer than expected: ${encoded.length} > ${finalOffset}`);
    return consumeAll ? decoded : [decoded, finalOffset];
}
function internalDeserializeLayout(layout, encoded, offset) {
    let decoded = {};
    for (const item of layout)
        [(item.omit ? {} : decoded)[item.name], offset] =
            deserializeLayoutItem(item, encoded, offset);
    return [decoded, offset];
}
function updateOffset(encoded, offset, size) {
    const newOffset = offset + size;
    if (newOffset > encoded.length)
        throw new Error(`encoded data is shorter than expected: ${encoded.length} < ${newOffset}`);
    return newOffset;
}
function deserializeUint(encoded, offset, size) {
    let value = 0n;
    for (let i = 0; i < size; ++i)
        value += BigInt(encoded[offset + i]) << BigInt((size - 1 - i) * 8);
    return [
        ((size > numberMaxSize) ? value : Number(value)),
        updateOffset(encoded, offset, size)
    ];
}
function deserializeLayoutItem(item, encoded, offset) {
    try {
        switch (item.binary) {
            case "object": {
                return internalDeserializeLayout(item.layout, encoded, offset);
            }
            case "array": {
                let ret = [];
                if (item.lengthSize !== undefined) {
                    const [length, newOffset] = deserializeUint(encoded, offset, item.lengthSize);
                    offset = newOffset;
                    for (let i = 0; i < length; ++i)
                        [ret[i], offset] = internalDeserializeLayout(item.layout, encoded, offset);
                }
                else {
                    while (offset < encoded.length)
                        [ret[ret.length], offset] = internalDeserializeLayout(item.layout, encoded, offset);
                }
                return [ret, offset];
            }
            case "bytes": {
                let newOffset;
                let fixedFrom;
                let fixedTo;
                if (item.custom !== undefined) {
                    if (item.custom instanceof Uint8Array)
                        fixedFrom = item.custom;
                    else if (item.custom.from instanceof Uint8Array) {
                        fixedFrom = item.custom.from;
                        fixedTo = item.custom.to;
                    }
                }
                if (fixedFrom !== undefined)
                    newOffset = updateOffset(encoded, offset, fixedFrom.length);
                else {
                    item = item;
                    if ("size" in item && item.size !== undefined)
                        newOffset = updateOffset(encoded, offset, item.size);
                    else if (item.lengthSize !== undefined) {
                        let length;
                        [length, offset] = deserializeUint(encoded, offset, item.lengthSize);
                        newOffset = updateOffset(encoded, offset, length);
                    }
                    else
                        newOffset = encoded.length;
                }
                const value = encoded.slice(offset, newOffset);
                if (fixedFrom !== undefined) {
                    checkUint8ArrayDeeplyEqual(fixedFrom, value);
                    return [fixedTo ?? fixedFrom, newOffset];
                }
                return [
                    item.custom !== undefined ? item.custom.to(value) : value,
                    newOffset
                ];
            }
            case "uint": {
                const [value, newOffset] = deserializeUint(encoded, offset, item.size);
                if (item.custom !== undefined) {
                    if (typeof item.custom === "number" || typeof item.custom === "bigint") {
                        checkUintEquals(item.custom, value);
                        return [item.custom, newOffset];
                    }
                    else if (typeof item.custom.from === "number" || typeof item.custom.from === "bigint") {
                        checkUintEquals(item.custom.from, value);
                        return [item.custom.to, newOffset];
                    }
                }
                return [
                    item.custom !== undefined ? item.custom.to(value) : value,
                    newOffset
                ];
            }
        }
    }
    catch (e) {
        e.message = `when deserializing item '${item.name}': ${e.message}`;
        throw e;
    }
}
//# sourceMappingURL=deserialize.js.map